/*
*  Arnold emulator (c) Copyright, Kevin Thacker 1995-2015
*
*  This file is part of the Arnold emulator source code distribution.
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include "cpcglob.h"
#include "ata.h"

/* LBA = ((cylinder * heads_per_cylinder + heads) * sectors_per_track) + sector - 1 */
void ata_lba_to_chs(ata_device *device, unsigned long LBA, unsigned short *C, unsigned char *H, unsigned char *S)
{
	/* iterate over sectors per cylinder, then each head in turn */
	*C = LBA / (device->Heads*device->SectorsPerTrack);
	*H = (LBA / device->SectorsPerTrack) % device->Heads;
	*S = (LBA % device->SectorsPerTrack) + 1;
}

void ata_mark_sector_exists(ata_device *device, int nSector)
{
	int nByte = nSector >> 3;
	int nBit = nSector & 0x07;
	device->SectorExistsMap[nByte] |= (1 << nBit);
}

BOOL ata_sector_exists(ata_device *device, int nSector)
{
	int nByte = nSector >> 3;
	int nBit = nSector & 0x07;
	if (device->SectorExistsMap[nByte] & (1 << nBit))
		return TRUE;

	return FALSE;
}

unsigned char *ata_get_sector_data(ata_device *device, int nSector)
{
	if (!ata_sector_exists(device, nSector))
	{
		/* mark as existing */
		ata_mark_sector_exists(device, nSector);
		/* allocate */
		device->SectorPointers[nSector] = malloc(ATA_SECTOR_SIZE);
		if (device->SectorPointers[nSector])
		{
			memset(device->SectorPointers[nSector], 0, ATA_SECTOR_SIZE);
		}
	}
	return device->SectorPointers[nSector];
}

void ata_chs_partition(ata_device *device, unsigned char *pCHS, int Sector)
{
	unsigned short C;
	unsigned char H;
	unsigned char S;
	ata_lba_to_chs(device, Sector, &C, &H, &S);
	pCHS[0] = H;
	pCHS[2] = C & 0x0ff;
	pCHS[1] = (S & 0x03f) | ((C >> 2) & 0x0c0);
}

void ata_setup_partition(ata_device *device, unsigned char *pPartition, int StartSector, int LengthSectors)
{
	pPartition[0] = 0x080;
	ata_chs_partition(device, &pPartition[1], StartSector);
	pPartition[4] = 0x01;
	ata_chs_partition(device, &pPartition[5], StartSector + LengthSectors);
	pPartition[8] = StartSector & 0x0ff;
	pPartition[9] = (StartSector>>8) & 0x0ff;
	pPartition[10] = (StartSector>>16) & 0x0ff;
	pPartition[11] = (StartSector>>24) & 0x00f;
	pPartition[12] = (LengthSectors & 0x0ff);
	pPartition[13] = ((LengthSectors>>8) & 0x0ff);
	pPartition[14] = ((LengthSectors >> 16) & 0x0ff);
	pPartition[15] = ((LengthSectors >> 24) & 0x00f);

}

void ata_init(ata_device *device, int nCapacity)
{
	/* need to set drq? */
	device->DeviceMaxSectors = nCapacity / ATA_SECTOR_SIZE;

	int SectorExistsMapSize;

	SectorExistsMapSize = device->DeviceMaxSectors >> 3;
	device->SectorExistsMap = (unsigned char *)malloc(SectorExistsMapSize);
	device->SectorPointers = (unsigned char **)malloc(sizeof(unsigned char *)*device->DeviceMaxSectors);
	memset(device->SectorPointers, 0, sizeof(unsigned char *)*device->DeviceMaxSectors);

	memset(device->SectorExistsMap, 0, SectorExistsMapSize);

//	device->SectorsPerTrack = 63;
//	device->Heads = 16;

	device->Heads = 1;
	device->SectorsPerTrack = 32;
	memset(device->DeviceIdentification, 0, sizeof(device->DeviceIdentification));

	if (device->DeviceMaxSectors >= MAX_CHS_SECTORS)
	{
		/* set LBA; correct?? */
		device->HeadRegister |= (1 << 6);
		device->HeadRegister |= (1 << 7);
		device->HeadRegister |= (1 << 5);
	}
	{
		unsigned short MaxC;
		unsigned char MaxH, MaxS;
		
		/* this doesn't work correctly it seems; returns less than heads... */
		ata_lba_to_chs(device, device->DeviceMaxSectors, &MaxC, &MaxH, &MaxS);
		MaxC++;
		MaxH++;
		MaxS++;
		/* TODO fill in remaining */

		/* total number of user addressable sectors (LBA mode) */
		device->DeviceIdentification[0] = (1 << 6);

		int NumberOfCylinders = 16383;
		if (device->DeviceMaxSectors < MAX_CHS_SECTORS)
		{
			NumberOfCylinders = MaxC+1;
		}
		device->DeviceIdentification[(1 * 2)] = (NumberOfCylinders & 0x0ff);
		device->DeviceIdentification[(1 * 2) + 1] = ((NumberOfCylinders >> 8) & 0x0ff);

		int NumberOfHeads = 16;
		if (device->DeviceMaxSectors < MAX_CHS_SECTORS)
		{
			NumberOfHeads = device->Heads;
		}
		device->DeviceIdentification[(3 * 2)] = NumberOfHeads & 0x0ff;
		device->DeviceIdentification[(3 * 2) + 1] = (NumberOfHeads >> 8) & 0x0ff;

		int NumberOfSectorsPerCylinder = 63;
		if (device->DeviceMaxSectors < MAX_CHS_SECTORS)
		{
			NumberOfSectorsPerCylinder = device->SectorsPerTrack;
		}
		device->DeviceIdentification[(6 * 2)] = NumberOfSectorsPerCylinder & 0x0ff;
		device->DeviceIdentification[(6 * 2) + 1] = (NumberOfSectorsPerCylinder >> 8) & 0x0ff;


		device->DeviceIdentification[(54 * 2)] = MaxC & 0x0ff;
		device->DeviceIdentification[(54 * 2) + 1] = (MaxC >> 8) & 0x0ff;
		device->DeviceIdentification[(55 * 2)] = MaxH & 0x0ff;
		device->DeviceIdentification[(56 * 2)] = MaxS & 0x0ff;

		const char *sIdentification = "Emulated IDE Harddisc";
		memset(&device->DeviceIdentification[10 * 2], '0', 20);
		memset(&device->DeviceIdentification[23 * 2], '0', 8);
		memset(&device->DeviceIdentification[27 * 2], ' ', 40);
		memcpy(&device->DeviceIdentification[27 * 2], sIdentification, strlen(sIdentification));
		{
			int i;
			/* bytes appeared to be reversed, at least they appear to be this way on BonnyDOS whereas the other values
			are not reversed */
			for ( i = 0; i < 20; i++)
			{
				unsigned char Byte = device->DeviceIdentification[(27 * 2) + (i << 1)];
				device->DeviceIdentification[(27 * 2) + (i << 1)] = device->DeviceIdentification[(27 * 2) + (i << 1) + 1];
				device->DeviceIdentification[(27 * 2) + (i << 1) + 1] = Byte;
			}
		}
			
		device->DeviceIdentification[(60 * 2) + 0] = device->DeviceMaxSectors & 0x0ff;
		device->DeviceIdentification[(60 * 2) + 1] = (device->DeviceMaxSectors >> 8) & 0x0ff;
		device->DeviceIdentification[(60 * 2) + 2] = (device->DeviceMaxSectors >> 16) & 0x0ff;
		device->DeviceIdentification[(60 * 2) + 3] = (device->DeviceMaxSectors >> 24) & 0x0f;
		device->DeviceIdentification[(80 * 2) + 0] = 14; /* ATA 14 supported */

	}

	{
		unsigned char *pMBR = ata_get_sector_data(device, 0);
		ata_setup_partition(device, &pMBR[0x01be], 1, device->DeviceMaxSectors - 1);
	}
}

void ata_finish(ata_device *device)
{
	int i;
	for (i = 0; i < device->DeviceMaxSectors; i++)
	{
		if (device->SectorPointers[i] != NULL)
		{
			free(device->SectorPointers[i]);
			device->SectorPointers[i] = NULL;
		}
	}
	free(device->SectorPointers);
	free(device->SectorExistsMap);
}

void ata_reset(ata_device *device)
{

	device->StatusRegister = 0;
	/* device is ready */
	device->StatusRegister |= DRDY;

	device->SectorsRemaining = 0;
	device->DataSize = 0;
	device->bTransferingSectors = FALSE;

}

unsigned char ata_get_status_register(ata_device *device)
{
	device->StatusRegister |= DRDY;
//	if ((device->DataSize != 0) && (device->SectorsRemaining != 0))
//	{
//		device->StatusRegister |= DRQ;
//	}

	/* alterate status register */
	return device->StatusRegister;
}

unsigned char ata_read_sector_data(ata_device *device)
{
	unsigned char Data = 0;
	if (ata_sector_exists(device, device->CurrentLBASectorIndex))
	{
		unsigned char *pSector = ata_get_sector_data(device, device->CurrentLBASectorIndex);
		Data = pSector[device->DataSize];
	}

	device->DataSize++;
	/* has data in sector remaining? */
	if (device->DataSize == ATA_SECTOR_SIZE)
	{
		device->SectorsRemaining--;
		if (device->SectorsRemaining == 0)
		{
			/* no more data */
			device->StatusRegister &= ~DRQ;
		}
		else
		{
			device->DataSize = 0;
			device->CurrentLBASectorIndex++;
			//printf("Read sector next %lu\n", device->CurrentLBASectorIndex);

			/* if sector is not valid indicate the error */
			if (device->CurrentLBASectorIndex >= device->DeviceMaxSectors)
			{
				device->StatusRegister |= ERR;
				device->StatusRegister &= ~DRQ;
				device->ErrorRegister |= ABRT;
				device->ErrorRegister |= IDNF;
			}
		}
	}

	return Data;
}

void ata_write_sector_data(ata_device *device, unsigned char Data)
{
	unsigned char *pSector = ata_get_sector_data(device, device->CurrentLBASectorIndex);
	pSector[device->DataSize] = Data;
	device->DataSize++;
	/* has data in sector remaining? */
	if (device->DataSize == ATA_SECTOR_SIZE)
	{
		device->SectorsRemaining--;
		if (device->SectorsRemaining == 0)
		{
			/* no more data */
			device->StatusRegister &= ~DRQ;
		}
		else
		{
			device->DataSize = 0;
			device->CurrentLBASectorIndex++;

			//printf("Write sector next %lu\n", device->CurrentLBASectorIndex);

			/* if sector is not valid indicate the error */
			if (device->CurrentLBASectorIndex >= device->DeviceMaxSectors)
			{
				device->StatusRegister |= ERR;
				device->StatusRegister &= ~DRQ;
				device->ErrorRegister |= ABRT;
				device->ErrorRegister |= IDNF;
			}
		}
	}
}


int ata_read(ata_device *device, int reg)
{
	switch (reg)
	{
	case ATA_ALTERNATE_STATUS_CS | ATA_ALTERNATE_STATUS_DA:
	{
		/* alternative status register */
		return ata_get_status_register(device);
	}

	case ATA_STATUS_CS | ATA_STATUS_DA:
	{
		/* status register */
		return ata_get_status_register(device);
	}


	case ATA_CYLINDER_NUMBER_HIGH_CS | ATA_CYLINDER_NUMBER_HIGH_DA:
		/* cylinder high register */
		return device->CylinderHigh;


	case ATA_CYLINDER_NUMBER_LOW_CS | ATA_CYLINDER_NUMBER_LOW_DA:
		/* drq and bsy should be 0 */
		/* cylinder high register */
		return device->CylinderLow;


	case ATA_DATA_REGISTER_CS | ATA_DATA_REGISTER_DA:
	{
		if (device->bTransferingSectors)
		{
			return ata_read_sector_data(device);
		}
		else
		{
			unsigned char Data = device->DeviceIdentification[device->DataSize];
			device->DataSize++;
			if (device->DataSize == ATA_SECTOR_SIZE)
			{
				device->StatusRegister &= ~DRQ;
			}
			return Data;
		}
	}
	break;

	case ATA_ERROR_CS | ATA_ERROR_DA:
		return device->ErrorRegister;

	case ATA_SECTOR_COUNT_CS | ATA_SECTOR_COUNT_DA:
		return device->SectorCountRegister;

	case ATA_SECTOR_NUMBER_CS | ATA_SECTOR_NUMBER_DA:
		return device->SectorNumberRegister;

	case ATA_HEAD_CS | ATA_HEAD_DA:
		return device->HeadRegister;

	default:
	{
		//printf("Unhandled!");
		break;

	}
	}

	return 0x0;
}

void ata_calc_initial_sector_index(ata_device *device)
{
	if (device->HeadRegister & (1 << 6))
	{
		device->CurrentLBASectorIndex =
			((device->HeadRegister & 0x0f) << 24) |
			((device->CylinderHigh & 0x0ff) << 16) |
			((device->CylinderLow & 0x0ff) << 8) |
			(device->SectorNumberRegister & 0x0ff);
		//printf("LBA %lu\n", device->CurrentLBASectorIndex);
	}
	else
	{
		unsigned short Cylinder =
			(device->CylinderLow & 0x0ff) |
			((device->CylinderHigh & 0x0ff) << 8);

		if (device->SectorNumberRegister == 0)
		{
			/* error; abort */
			device->ErrorRegister |= ABRT;
		}
		else
		{
			device->CurrentLBASectorIndex =
				(((Cylinder * device->Heads) + (device->HeadRegister & 0x0f)) * device->SectorsPerTrack) +
				((device->SectorNumberRegister & 0x0ff) - 1);
			//printf("Cylinder %d Head %d Sector %d\n", Cylinder, device->HeadRegister & 0x0f, device->SectorNumberRegister);
		}
	}
}

void ata_write(ata_device *device, int reg, int data)
{	
	switch (reg)
	{
		case CS1_REG_BIT|DA2_REG_BIT|DA1_REG_BIT:
		{
			/* device control register */
			device->DeviceControl = data;
		}
		break;

		case ATA_COMMAND_CS | ATA_COMMAND_DA:
		{
			/* command register */

			/* clear interrupt */

			/* drq and bsy should be 0 */


			device->CommandRegister = data;
			device->StatusRegister |= BSY;
			device->StatusRegister &= ~ERR;
			device->ErrorRegister = 0;

			switch (device->CommandRegister)
			{
				/* write sectors with retry */
			case WRITE_MULTIPLE:
			case WRITE_SECTORS_WITH_RETRIES:
			case WRITE_SECTORS_WITHOUT_RETRIES:
			{
				ata_calc_initial_sector_index(device);
				//printf("Write sector %lu\n", device->CurrentLBASectorIndex);

				if (device->CurrentLBASectorIndex >= device->DeviceMaxSectors)
				{
					device->ErrorRegister |= ABRT;
					device->ErrorRegister |= IDNF;
				}

				if (device->ErrorRegister != 0)
				{
					device->StatusRegister |= ERR;
					device->StatusRegister &= ~BSY;
				}
				else
				{

					unsigned long nSectors = device->SectorCountRegister;
					if (nSectors == 0)
					{
						nSectors = 256;
					}
					device->DataSize = 0;
					device->SectorsRemaining = nSectors;
					device->bTransferingSectors = TRUE;
					device->StatusRegister |= DRQ;
					device->StatusRegister &= ~BSY;
				}
			}
			break;

						/* write sectors with retry */
			case WRITE_SECTORS_EXT:
			{
				if ((device->HeadRegister & (1 << 6))==0)
				{
					device->ErrorRegister |= ABRT;
				}
				else
				{
					device->CurrentLBASectorIndex =
						((device->HeadRegister & 0x0f) << 24) |
						((device->CylinderHigh & 0x0ff) << 16) |
						((device->CylinderLow & 0x0ff) << 8) |
						(device->SectorNumberRegister & 0x0ff);
					//printf("LBA %lu\n", device->CurrentLBASectorIndex);
				}
				//printf("Write sector %lu\n", device->CurrentLBASectorIndex);

				if (device->CurrentLBASectorIndex >= device->DeviceMaxSectors)
				{
					device->ErrorRegister |= ABRT;
					device->ErrorRegister |= IDNF;
				}

				if (device->ErrorRegister != 0)
				{
					device->StatusRegister |= ERR;
					device->StatusRegister &= ~BSY;
				}
				else
				{

					unsigned long nSectors = device->SectorCountRegister;
					if (nSectors == 0)
					{
						nSectors = 65536;
					}
					device->DataSize = 0;
					device->SectorsRemaining = nSectors;
					device->bTransferingSectors = TRUE;
					device->StatusRegister |= DRQ;
					device->StatusRegister &= ~BSY;
				}
			}
			break;
			
			/* read sectors with retry */
			case READ_MULTIPLE:
			case READ_SECTORS_WITH_RETRIES:
			case READ_SECTORS_WITHOUT_RETRIES:
			{
				ata_calc_initial_sector_index(device);
				//printf("Read sector %lu\n", device->CurrentLBASectorIndex);

				if (device->CurrentLBASectorIndex >= device->DeviceMaxSectors)
				{
					device->ErrorRegister |= ABRT;
					device->ErrorRegister |= IDNF;
				}

				if (device->ErrorRegister != 0)
				{
					device->StatusRegister |= ERR;
					device->StatusRegister &= ~BSY;
				}
				else
				{

					unsigned long nSectors = device->SectorCountRegister;
					if (nSectors == 0)
					{
						nSectors = 256;
					}
					device->DataSize = 0;
					device->SectorsRemaining = nSectors;
					device->bTransferingSectors = TRUE;
					device->StatusRegister |= DRQ;
					device->StatusRegister &= ~BSY;
				}
			}
			break;
			
			case READ_SECTORS_EXT:
			{
				if ((device->HeadRegister & (1 << 6))==0)
				{
					device->ErrorRegister |= ABRT;
				}
				else
				{
					device->CurrentLBASectorIndex =
						((device->HeadRegister & 0x0f) << 24) |
						((device->CylinderHigh & 0x0ff) << 16) |
						((device->CylinderLow & 0x0ff) << 8) |
						(device->SectorNumberRegister & 0x0ff);
					//printf("LBA %lu\n", device->CurrentLBASectorIndex);
				}
				
				//printf("Read sector %lu\n", device->CurrentLBASectorIndex);

				if (device->CurrentLBASectorIndex >= device->DeviceMaxSectors)
				{
					device->ErrorRegister |= ABRT;
					device->ErrorRegister |= IDNF;
				}

				if (device->ErrorRegister != 0)
				{
					device->StatusRegister |= ERR;
					device->StatusRegister &= ~BSY;
				}
				else
				{

					unsigned long nSectors = device->SectorCountRegister;
					if (nSectors == 0)
					{
						nSectors = 65536;
					}
					device->DataSize = 0;
					device->SectorsRemaining = nSectors;
					device->bTransferingSectors = TRUE;
					device->StatusRegister |= DRQ;
					device->StatusRegister &= ~BSY;
				}
			}
			break;


			case IDENTIFY_DEVICE:
			{
				device->bTransferingSectors = FALSE;
				device->DataSize = 0;
				device->StatusRegister |= DRQ;
				device->StatusRegister &= ~BSY;
				device->SectorsRemaining = 0;

				if (device->DeviceMaxSectors >= MAX_CHS_SECTORS)
				{
					/* set LBA; correct?? */
					device->HeadRegister |= (1 << 6);
					device->HeadRegister |= (1 << 7);
					device->HeadRegister |= (1 << 5);
				}

			}
			break;

			case READ_NATIVE_MAX_ADDRESS:
			{
				device->StatusRegister &= ~BSY;
				device->StatusRegister &= ~DRQ;
				device->ErrorRegister = 0;

				/* return the maximum sector address as LBA in the registers */
				device->SectorNumberRegister = device->DeviceMaxSectors & 0x0ff;
				device->CylinderLow = (device->DeviceMaxSectors>>8) & 0x0ff;
				device->CylinderHigh = (device->DeviceMaxSectors>>16) & 0x0ff;
				device->HeadRegister &= ~0x0f;
				device->HeadRegister |= (1 << 6);	// LBA
				device->HeadRegister |= (device->DeviceMaxSectors >> 24) & 0x0f;
			}
			break;

			case NOP_COMMAND:
				device->StatusRegister |= ERR;
				device->StatusRegister &= ~BSY;
				device->StatusRegister &= ~DRQ;
				device->ErrorRegister |= ABRT;
				break;

			default:
				device->StatusRegister |= ERR;
				device->StatusRegister &= ~BSY;
				device->StatusRegister &= ~DRQ;
				device->ErrorRegister |= ABRT;
				break;
			}
		}
		break;

		case ATA_CYLINDER_NUMBER_HIGH_CS | ATA_CYLINDER_NUMBER_HIGH_DA:
		{
			/* drq and bsy should be 0 */
			/* cylinder high register */
			device->CylinderHigh = data;
		}
		break;

		case ATA_CYLINDER_NUMBER_LOW_CS | ATA_CYLINDER_NUMBER_LOW_DA:
		{
			/* drq and bsy should be 0 */
			/* cylinder high register */
			device->CylinderLow = data;
		}
		break;


		case ATA_DATA_REGISTER_CS | ATA_DATA_REGISTER_DA:
		{
			if (device->bTransferingSectors)
			{
				ata_write_sector_data(device, data);
			}
		}
		break;

		case ATA_HEAD_CS | ATA_HEAD_DA:
		{
			/* bit 6 = LBA */
			/* bit 4 = device */
			device->HeadRegister = data;
		}
		break;



		case ATA_FEATURES_CS | ATA_FEATURES_DA:
		{
			device->FeaturesRegister = data;
		}
		break;


		case ATA_SECTOR_COUNT_CS | ATA_SECTOR_COUNT_DA:
		{
			device->SectorCountRegister = data;
		}
		break;


		case ATA_SECTOR_NUMBER_CS | ATA_SECTOR_NUMBER_DA:
		{
			device->SectorNumberRegister = data;
		}
		break;
		
		default:
		{
			//printf("Unhandled!");
			break;

		}

	}
}

