/*
 *  Arnold emulator (c) Copyright, Kevin Thacker 1995-2015
 *
 *  This file is part of the Arnold emulator source code distribution.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/********************/
/* snapshot support */
/********************/

/* The data common to snapshots 1,2,3 are in CPCEMU compatible form: */
/* - ink values are stored with bits 5,6,7 = 0 */
/* - PPI Control is the I/O state and mode of the ports and not the last value written to this port */
/* - PPI port C is  the last value written to this port */
/* - PPI port B is the inputs to this port */
/* - PPI port A is the inputs to this port */
/* - ram configuration does not have bit 7 and bit 6 set to 1 */

/* Version 3 has the additional chunks as written by No$CPC v1.8 */
/*
 TODO: Fix for all computers
 TODO: Fix uncompressed V3 causes crash
 */
#include "headers.h"

#include "snapshot.h"
#include "cpc.h"
#include "snapv3.h"
#include "fdi.h"
#include "i8255.h"
#include "psg.h"
#include "crtc.h"
#include "asic.h"
#include "kcc.h"
#include "printer.h"
#include "garray.h"
#include "aleste.h"

#ifndef max
#define max(a,b) (a>b) ? a : b
#endif

#ifndef min
#define min(a,b) (a<b) ? a : b
#endif
extern AY_3_8912 OnBoardAY;
extern char *Z80MemoryBase;
//extern char *Amstrad_ExtraRam;

/* get a register pair from memory */
#define SNAPSHOT_GET_REGISTER_PAIR(ptr,offset) \
	((ptr[offset+1] & 0x0ff)<<8) | (ptr[offset+0] & 0x0ff);

/* get a register from memory */
#define SNAPSHOT_GET_REGISTER(ptr,offset) \
	(ptr[offset] & 0x0ff)

/* put a register pair to memory */
#define SNAPSHOT_PUT_REGISTER_PAIR(ptr,offset,data) \
	ptr[offset+0] = data & 0x0ff; \
	ptr[offset+1] = (data>>8) & 0x0ff

/* put a register to memory */
#define SNAPSHOT_PUT_REGISTER(ptr,offset,data) \
	ptr[offset+0] = data & 0x0ff

#define SNAPSHOT_CRTC_FLAGS_HSYNC_STATE (1<<0)
#define SNAPSHOT_CRTC_FLAGS_VSYNC_STATE (1<<1)
#define SNAPSHOT_CRTC_FLAGS_VADJ_STATE (1<<7)

static SNAPSHOT_OPTIONS Options;

void SnapshotSettings_SetVersion(int nVersion)
{
	Options.Version = nVersion;
}

int SnapshotSettings_GetVersion(void)
{
	return Options.Version;
}

BOOL SnapshotSettings_GetCompressed(void)
{
	return Options.bCompressed;
}

void SnapshotSettings_SetCompressed(BOOL bState)
{
	Options.bCompressed = bState;
}

/* insert snapshot */
int Snapshot_Insert(const unsigned char *pSnapshot, const unsigned long SnapshotLength)
{
	BOOL Status = ARNOLD_STATUS_UNRECOGNISED;
	SNAPSHOT_HEADER *pSnapshotHeader;

	/* failed so quit */
	if (pSnapshot != NULL)
	{
		pSnapshotHeader = (SNAPSHOT_HEADER *)pSnapshot;

		/* snapshot length must be at least the size of snapshot header */
		if (SnapshotLength > sizeof(SNAPSHOT_HEADER))
		{
			/* is snapshot header text present? */
			if (memcmp(pSnapshotHeader, SNAPSHOT_HEADER_TEXT, 8) == 0)
			{
				/* header present */
				printf("Snapshot version: %d\n", pSnapshotHeader->Version);
				/* check version */
				if ((pSnapshotHeader->Version >= 1) && (pSnapshotHeader->Version <= 3))
				{
					char    *pSnapshotData;
					int i;

					/* size of actual data in snapshot */
					int ActualMemSizeInSnapshot = SnapshotLength - sizeof(SNAPSHOT_HEADER);

					/* mem size claimed to be in snapshot */
					int MemSizeInSnapshot = ((pSnapshotHeader->MemSizeLow & 0x0ff) | ((pSnapshotHeader->MemSizeHigh & 0x0ff) << 8)) << 10;

					/* we will use the larger value to determine the memory expansion we will enable */
					int MemSizeMax = max(ActualMemSizeInSnapshot, MemSizeInSnapshot);

					/* this is the size of memory chosen when enabling expansions */
					/*										unsigned long MemSizeChosen = 64*1024; */

					/* ram config required */
					//									unsigned long RamConfig = 0;

					printf("Actual size of snapshot data: %d\n", ActualMemSizeInSnapshot);
					printf("Mem size reported by snapshot: %d\n", MemSizeInSnapshot);
					printf("Mem size max: %d\n", MemSizeMax);
#if 0
					if (MemSizeMax<=(64*1024))
					{
						/* no extra ram */
					}
					else
						if (MemSizeMax<=(128*1024))
						{
							/* assume standard 64k ram expansion */
							RamConfig |= CPC_RAM_CONFIG_64K_RAM;
							MemSizeChosen = 128*1024;
						}
						else
							if (MemSizeMax<=(256*1024))
							{
								/* assume standard 256k ram */
								RamConfig |= CPC_RAM_CONFIG_256K_RAM;
								MemSizeChosen = 256*1024;
							}
							else
								if (MemSizeMax<=(512*1024))
								{
									/* assume 256k RAM and 256K silicon disc connected */
									RamConfig |= CPC_RAM_CONFIG_256K_RAM|CPC_RAM_CONFIG_256K_SILICON_DISK;
									MemSizeChosen = 512*1024;
								}

					printf("Mem size chosen: %lu\n", MemSizeChosen);
					printf("TODO: Mem Size correctly chosen. Needs to enable devices in future build!\n");
					/* force RAM config; ensuring existing ram config is maintained */
					//CPC_SetRamConfig(CPC_GetRamConfig() | RamConfig);
					// TODO!

#endif

					/* file must be at least the size of the header */
					if (SnapshotLength >= sizeof(SNAPSHOT_HEADER))
					{
						/* Appears snapshot is valid. Setup Z80 and copy memory */

						unsigned long CopySize;
						int CPCType;
						int Register;

						/**** INIT Z80 ****/
						/* setup Z80 */
						char *pRegs = (char *)pSnapshotHeader + 0x011;

						printf("Performing restart\n");
						/* reset CPC */
						Computer_RestartReset();
						/* get CPC type */
						CPCType = ((char *)pSnapshotHeader)[0x06d];

						printf("Snapshot CPCType (not reliable): %d\n", CPCType);

						if (
							(pSnapshotHeader->Version == 1) ||
							(pSnapshotHeader->Version == 2))
						{
							/* version 1 and 2 did not know about CPC+ so do not allow them
							to be set */

							/* set CPC type accordingly */
							switch (CPCType)
							{
							case 0:
							{
								CPC_SetHardware(CPC_HW_CPC);
								/*															CPC_SetCPCType(CPC_TYPE_CPC464); */
							}
							break;

							case 1:
							{
								CPC_SetHardware(CPC_HW_CPC);
								/*															CPC_SetCPCType(CPC_TYPE_CPC664); */
								/*															CPC_SetCPCType(CPC_TYPE_CPC664); */
							}
							break;

							/* for unknown or other values, assume cpc6128 */
							default:
							case 2:
							{
								CPC_SetHardware(CPC_HW_CPC);
								/*															CPC_SetCPCType(CPC_TYPE_CPC6128); */
							}
							break;

							}
						}
						else
							if (pSnapshotHeader->Version == 3)
							{
								printf("Should we guess + by the Plus chunk existing!?\n");

								/* version 3 understands CPC+ */
								/* allow them to be set */

								/* set CPC type accordingly */
								switch (CPCType)
								{
								case 0:
								{
									CPC_SetHardware(CPC_HW_CPC);
									/*															CPC_SetCPCType(CPC_TYPE_CPC464); */
								}
								break;

								case 1:
								{
									CPC_SetHardware(CPC_HW_CPC);
									/*															CPC_SetCPCType(CPC_TYPE_CPC664); */
								}
								break;

								case 2:
								{
									CPC_SetHardware(CPC_HW_CPC);
									/*															CPC_SetCPCType(CPC_TYPE_CPC6128); */

								}
								break;


								/* 6 is GX4000 */
								case 6:
								{
									ASIC_SetGX4000(TRUE);
									CPC_SetHardware(CPC_HW_CPCPLUS);
								}
								break;

								/* 4 is 464Plus */
								/* 5 is 6128Plus */
								case 4:
								case 5:
								{
									ASIC_SetGX4000(FALSE);
									CPC_SetHardware(CPC_HW_CPCPLUS);
								}
								break;

								/* for unknown etc assume 6128+ */
								default:
								{
									CPC_SetHardware(CPC_HW_CPCPLUS);
									/*															CPC_SetCPCType(CPC_TYPE_6128PLUS);*/
								}
								break;
								}
							}

						if (CPC_GetHardware() == CPC_HW_CPCPLUS)
						{
							ASIC_SetR128(FALSE);
							//				if (RamConfig & CPC_RAM_CONFIG_64K_RAM)
							//				{
							//					ASIC_SetR128(TRUE);
							//				}
						}
						else
							if (CPC_GetHardware() == CPC_HW_CPC)
							{
								/* if 6128.. */
								//			PAL_Uninstall();
								//			if (RamConfig & CPC_RAM_CONFIG_64K_RAM)
								//			{
								//				PAL_Install();
								//			}
							}
						// Aleste doesn't need any ram configuration it has enough ram
						// KC Compact needs a ram expansion enabling


						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 0);
						CPU_SetReg(CPU_AF, Register);
						printf("Snapshot Z80 AF: &%04x\n", Register);



						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 2);
						CPU_SetReg(CPU_BC, Register);
						printf("Snapshot Z80 BC: &%04x\n", Register);


						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 4);
						CPU_SetReg(CPU_DE, Register);
						printf("Snapshot Z80 DE: &%04x\n", Register);

						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 6);
						CPU_SetReg(CPU_HL, Register);
						printf("Snapshot Z80 HL: &%04x\n", Register);


						CPU_SetReg(CPU_R, SNAPSHOT_GET_REGISTER(pRegs, 8));
						printf("Snapshot Z80 R: &%02x\n", SNAPSHOT_GET_REGISTER(pRegs, 8));


						CPU_SetReg(CPU_I, SNAPSHOT_GET_REGISTER(pRegs, 9));
						printf("Snapshot Z80 I: &%02x\n", SNAPSHOT_GET_REGISTER(pRegs, 9));


						CPU_SetReg(CPU_IFF1, pRegs[10] & 1);
						printf("Snapshot Z80 IFF1: &%04x\n", pRegs[10] & 1);


						CPU_SetReg(CPU_IFF2, pRegs[11] & 1);
						printf("Snapshot Z80 IFF2: &%04x\n", pRegs[11] & 1);

						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 12);
						CPU_SetReg(CPU_IX, Register);
						printf("Snapshot Z80 IX: &%04x\n", Register);


						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 14);
						CPU_SetReg(CPU_IY, Register);
						printf("Snapshot Z80 IY: &%04x\n", Register);


						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 16);
						CPU_SetReg(CPU_SP, Register);
						printf("Snapshot Z80 SP: &%04x\n", Register);


						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 18);
						CPU_SetReg(CPU_PC, Register);

						printf("Snapshot Z80 PC: &%04x\n", Register);

						Register = pRegs[20];

						if ((Register < 0) || (Register>2))
						{
							/* IM 1 is better for CPC */
							Register = 1;
						}
						printf("Snapshot Z80 IM: &%04x\n", Register);

						CPU_SetReg(CPU_IM, Register);

						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 21);
						CPU_SetReg(CPU_AF2, Register);
						printf("Snapshot Z80 AF': &%04x\n", Register);

						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 23);
						CPU_SetReg(CPU_BC2, Register);
						printf("Snapshot Z80 BC': &%04x\n", Register);

						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 25);
						CPU_SetReg(CPU_DE2, Register);
						printf("Snapshot Z80 DE': &%04x\n", Register);

						Register = SNAPSHOT_GET_REGISTER_PAIR(pRegs, 27);
						CPU_SetReg(CPU_HL2, Register);
						printf("Snapshot Z80 HL': &%04x\n", Register);


						switch (CPC_GetHardware())
						{
						case CPC_HW_CPC:
						{
							/**** GATE ARRAY ****/
							/* initialise colour palette */
							for (i = 0; i < 17; i++)
							{
								unsigned char HwColourIndex = ((char *)pSnapshotHeader)[0x02f + i];

								/* pen select */
								GateArray_Write(i);

								/* write colour for pen */
								GateArray_Write((HwColourIndex & 0x01f) | 0x040);
								printf("GA Pen %d Colour &%02x\n", i, (HwColourIndex & 0x01f) | 0x040);

							}

							/* pen select */
							GateArray_Write(((char *)pSnapshotHeader)[0x02e] & 0x01f);
							printf("GA Selected Pen %d\n", ((char *)pSnapshotHeader)[0x02e] & 0x01f);

							/* mode and rom configuration select */
							GateArray_Write((((char *)pSnapshotHeader)[0x040] & 0x03f) | 0x080);
							printf("GA MREM %02x\n", (((char *)pSnapshotHeader)[0x040] & 0x03f) | 0x080);
						}
						break;


						case CPC_HW_CPCPLUS:
						{
							/**** GATE ARRAY inside ASIC ****/
							/* initialise colour palette */
							for (i = 0; i < 17; i++)
							{
								unsigned char HwColourIndex = ((char *)pSnapshotHeader)[0x02f + i];

								/* pen select */
								ASIC_GateArray_Write(i);

								/* write colour for pen */
								ASIC_GateArray_Write((HwColourIndex & 0x01f) | 0x040);
								printf("GA Pen %d Colour &%02x\n", i, (HwColourIndex & 0x01f) | 0x040);

							}

							/* pen select */
							ASIC_GateArray_Write(((char *)pSnapshotHeader)[0x02e] & 0x01f);
							printf("GA Selected Pen %d\n", ((char *)pSnapshotHeader)[0x02e] & 0x01f);

							/* mode and rom configuration select */
							ASIC_GateArray_Write((((char *)pSnapshotHeader)[0x040] & 0x03f) | 0x080);
							printf("GA MREM &%02x\n", (((char *)pSnapshotHeader)[0x040] & 0x03f) | 0x080);

						}
						break;

						case CPC_HW_KCCOMPACT:
						{
							/**** KCC hardware ****/
							/* initialise colour palette */
							for (i = 0; i < 17; i++)
							{
								unsigned char HwColourIndex = ((char *)pSnapshotHeader)[0x02f + i];

								/* pen select */
								KCC_GA_Write(i);

								/* write colour for pen */
								KCC_GA_Write((HwColourIndex & 0x01f) | 0x040);
								printf("(KCC) GA Pen %d Colour &%02x\n", i, (HwColourIndex & 0x01f) | 0x040);

							}

							/* pen select */
							KCC_GA_Write(((char *)pSnapshotHeader)[0x02e] & 0x01f);
							printf("(KCC) GA Selected Pen %d\n", ((char *)pSnapshotHeader)[0x02e] & 0x01f);

							/* mode and rom configuration select */
							KCC_GA_Write((((char *)pSnapshotHeader)[0x040] & 0x03f) | 0x080);
							printf("(KCC) GA MREM %02x\n", (((char *)pSnapshotHeader)[0x040] & 0x03f) | 0x080);

						}
						break;

						case CPC_HW_ALESTE:
						{
							printf("Aleste will not restore colours! FIXME!");

						}
						break;

						default:
						{
							printf("Colours not restored!!!\n");
						}
						break;

						}

						/* ram configuration select */
						/* TODO Fix */
						/*    PAL_WriteConfig((((char *)pSnapshotHeader)[0x041] & 0x03f) | 0x0c0); */

						/**** CRTC ****/
						/* initialise CRTC */
						for (i = 0; i < 18; i++)
						{
							unsigned char CRTCRegData = ((unsigned char *)pSnapshotHeader)[0x043 + i];

							CRTC_RegisterSelect(i);

							CRTC_WriteData(CRTCRegData & 0x0ff);
							printf("CRTC Register %d Data &%02x\n", i, CRTCRegData);
						}

						/* select CRTC register */
						CRTC_RegisterSelect(((char *)pSnapshotHeader)[0x042]);
						printf("CRTC Selected Register %d\n", ((char *)pSnapshotHeader)[0x042]);

						/**** ROM INDEX ****/
						/* TODO: select expansion rom*/
						//ROM_SetSelectedROMIndex(((char *)pSnapshotHeader)[0x055]);
						printf("Upper ROM %02x - FIXME!!!\n", ((char *)pSnapshotHeader)[0x055]);


						/**** PSG ****/
						/* setup PSG registers */
						for (i = 0; i < 16; i++)
						{
							unsigned char PSGRegData = ((unsigned char *)pSnapshotHeader)[0x05b + i];

							/* !!!! */
							PSG_RegisterSelect(&OnBoardAY, i);

							/* !!!!! */
							PSG_WriteData(&OnBoardAY, PSGRegData);
							printf("On-board AY Register %d Data &%02x\n", i, PSGRegData);
						}

						/* !!!! */
						/* select PSG register */
						PSG_RegisterSelect(&OnBoardAY, ((char *)pSnapshotHeader)[0x05a]);
						printf("On-board AY Register selected %d\n", ((char *)pSnapshotHeader)[0x05a]);

						switch (CPC_GetHardware())
						{
						case CPC_HW_CPCPLUS:
						{
							/**** ASIC PPI ****/

							/* set control */
							ASIC_PPI_WriteControl(((char *)pSnapshotHeader)[0x059]);
							printf("(Plus) PPI Control &%02x\n", ((char *)pSnapshotHeader)[0x059]);

							/* these depend on input/output! */
							/* set port A data */
							ASIC_PPI_SetPortDataFromSnapshot(0, ((char *)pSnapshotHeader)[0x056]);
							printf("(Plus) PPI Port A Data &%02x\n", ((char *)pSnapshotHeader)[0x056]);
							/* set port B data */
							ASIC_PPI_SetPortDataFromSnapshot(1, ((char *)pSnapshotHeader)[0x057]);
							printf("(Plus) PPI Port B Data &%02x\n", ((char *)pSnapshotHeader)[0x057]);
							/* set port C data */
							ASIC_PPI_SetPortDataFromSnapshot(2, ((char *)pSnapshotHeader)[0x058]);
							printf("(Plus) PPI Port C Data &%02x\n", ((char *)pSnapshotHeader)[0x058]);

							PSG_SetBDIRState(&OnBoardAY, ((char *)pSnapshotHeader)[0x058] & (1 << 7));
							PSG_SetBC1State(&OnBoardAY, ((char *)pSnapshotHeader)[0x058] & (1 << 6));
							PSG_RefreshState(&OnBoardAY);

						}
						break;

						default:
						{


							/**** PPI ****/

							/* set control */
							PPI_WriteControl(((char *)pSnapshotHeader)[0x059]);
							printf("PPI Control &%02x\n", ((char *)pSnapshotHeader)[0x059]);

							/* set port A data */
							PPI_SetPortDataFromSnapshot(0, ((char *)pSnapshotHeader)[0x056]);
							printf("PPI Port A data &%02x\n", ((char *)pSnapshotHeader)[0x056]);
							/* set port B data */
							PPI_SetPortDataFromSnapshot(1, ((char *)pSnapshotHeader)[0x057]);
							printf("PPI Port B data &%02x\n", ((char *)pSnapshotHeader)[0x057]);
							/* set port C data */
							PPI_SetPortDataFromSnapshot(2, ((char *)pSnapshotHeader)[0x058]);
							printf("PPI Port C data &%02x\n", ((char *)pSnapshotHeader)[0x058]);

							PSG_SetBDIRState(&OnBoardAY, ((char *)pSnapshotHeader)[0x058] & (1 << 7));
							PSG_SetBC1State(&OnBoardAY, ((char *)pSnapshotHeader)[0x058] & (1 << 6));
							PSG_RefreshState(&OnBoardAY);

						}
						break;
						}

						if (pSnapshotHeader->Version == 3)
						{
							unsigned long CRTC_Flags;
							int CRTCType;
							CRTC_INTERNAL_STATE *pCRTC_State;

							pCRTC_State = CRTC_GetInternalState();

							FDI_SetMotorState(((char *)pSnapshotHeader)[0x09c] & 0x01);
							printf("Disc motor %d\n", ((char *)pSnapshotHeader)[0x09c] & 0x01);

							Printer_Write7BitData(((char *)pSnapshotHeader)[0x0a1]);
							printf("Printer data%d\n", ((char *)pSnapshotHeader)[0x0a1]);
							Printer_SetStrobeState(((((char *)pSnapshotHeader)[0x0a1] & 0x080) != 0));

							CRTCType = ((char *)pSnapshotHeader)[0x0a4];

							if ((CRTCType < 0) || (CRTCType >= 5))
							{
								CRTCType = 0;
							}
							printf("CRTC type %d\n", ((char *)pSnapshotHeader)[0x0a4]);

							CRTC_SetType(CRTCType);

							pCRTC_State->HCount = ((char *)pSnapshotHeader)[0x0a9] & 0x0ff;
							printf("CRTC HC %d\n", ((char *)pSnapshotHeader)[0x0a9] & 0x0ff);
							pCRTC_State->LineCounter = ((char *)pSnapshotHeader)[0x0ab] & 0x07f;
							printf("CRTC LC %d\n", ((char *)pSnapshotHeader)[0x0ab] & 0x07f);
							pCRTC_State->RasterCounter = ((char *)pSnapshotHeader)[0x0ac] & 0x01f;
							printf("CRTC RA %d\n", ((char *)pSnapshotHeader)[0x0ac] & 0x01f);

							CRTC_Flags = (((char *)pSnapshotHeader)[0x0b0] & 0x0ff) | ((((char *)pSnapshotHeader)[0x0b1] & 0x0ff) << 8);

							pCRTC_State->CRTC_Flags &= ~CRTC_VS_FLAG;
							if (CRTC_Flags & SNAPSHOT_CRTC_FLAGS_VSYNC_STATE)
							{
								pCRTC_State->VerticalSyncCount = ((char *)pSnapshotHeader)[0x0af];
								pCRTC_State->CRTC_Flags |= CRTC_VS_FLAG;
							}

							pCRTC_State->CRTC_Flags &= ~CRTC_HS_FLAG;
							if (CRTC_Flags & SNAPSHOT_CRTC_FLAGS_HSYNC_STATE)
							{
								pCRTC_State->HorizontalSyncCount = ((char *)pSnapshotHeader)[0x0ae];
								pCRTC_State->CRTC_Flags |= CRTC_HS_FLAG;
							}

							pCRTC_State->CRTC_Flags &= ~CRTC_VADJ_FLAG;
							if (CRTC_Flags & SNAPSHOT_CRTC_FLAGS_VADJ_STATE)
							{
								pCRTC_State->VertAdjustCount = ((char *)pSnapshotHeader)[0x0ad];
								pCRTC_State->CRTC_Flags |= CRTC_VADJ_FLAG;
							}

							switch (CPC_GetHardware())
							{

							case CPC_HW_CPC:
							{
								GateArray_SetInterruptLineCount(((char *)pSnapshotHeader)[0x0b3]);
								printf("GA ILC %d\n", ((char *)pSnapshotHeader)[0x0b3]);


								// TODO:

								//           GateArray_SetVsyncSynchronisationCount(((char *)pSnapshotHeader)[0x0b2]);
								//printf("GA VSYNC SYNC COUNT %d\n", ((char *)pSnapshotHeader)[0x0b2]);
							}
							break;

							case CPC_HW_CPCPLUS:
							{
								ASIC_GateArray_SetInterruptLineCount(((char *)pSnapshotHeader)[0x0b3]);
								printf("ASIC ILC %d\n", ((char *)pSnapshotHeader)[0x0b3]);
								// TODO:

								//ASIC_GateArray_SetVsyncSynchronisationCount( ( (char *) pSnapshotHeader )[ 0x0b2 ] );
								//printf("ASIC VSYNC SYNC COUNT %d\n", ((char *)pSnapshotHeader)[0x0b2]);
							}
							break;

							case CPC_HW_KCCOMPACT:
							{
								printf("(KCC) Interrupt line count not set! FIXME!\n");

							}
							break;

							case CPC_HW_ALESTE:
							{
								printf("(Aleste) Interrupt line count not set! FIXME!\n");



							}
							break;

							default:
							{
								printf("Interrupt line count not set! FIXME!\n");

							}
							break;

							}

							printf("Z80 IRQ input %d\n", (((char *)pSnapshotHeader)[0x0b4] & 0x01));

							CPU_SetINTState((((char *)pSnapshotHeader)[0x0b4] & 0x01));
						}


						/**** INIT RAM ****/
						pSnapshotData = ((char *)pSnapshotHeader + sizeof(SNAPSHOT_HEADER));

						/* size of main ram to copy. */
						CopySize = min((SnapshotLength - sizeof(SNAPSHOT_HEADER)), 64 * 1024);
						printf("CopySize (base 64k): %lu\n", CopySize);

						memcpy(Z80MemoryBase, pSnapshotData, CopySize);

						CopySize = SnapshotLength - sizeof(SNAPSHOT_HEADER) - CopySize;

						/* now if there is extra data copy it in */
						if (CopySize != 0)
						{
#if 0
							/* extra ram setup */
							if (Amstrad_ExtraRam!=NULL)
							{
								unsigned long Length = min(CopySize, MemSizeChosen);

								/* copy data into it */
								memcpy(Amstrad_ExtraRam, pSnapshotData + (64<<10), Length);

								printf("Extra RAM Size: %lu\n", Length);

							}
#endif
						}


						Status = ARNOLD_STATUS_OK;

						if (pSnapshotHeader->Version == 3)
						{
							/* must be at least enough data for a chunk header after the memory dump */
							if (SnapshotLength > (MemSizeInSnapshot + sizeof(SNAPSHOT_HEADER) + sizeof(RIFF_CHUNK)))
							{
								/* some chunks after the data */
								RIFF_CHUNK *pChunk = (RIFF_CHUNK *)((char *)pSnapshotHeader + sizeof(SNAPSHOT_HEADER) + MemSizeInSnapshot);
								unsigned long SizeRemaining;

								SizeRemaining = SnapshotLength - MemSizeInSnapshot - sizeof(SNAPSHOT_HEADER);

								while (SizeRemaining != 0)
								{
									int ChunkSize;

									/* get length of this chunk's data in bytes */
									ChunkSize = Riff_GetChunkLength(pChunk);

									if (ChunkSize == 0)
										break;

									/* chunk size bad? */
									if (ChunkSize > (SizeRemaining - sizeof(RIFF_CHUNK)))
									{
										/* size of chunk is greater than size of data remaining */
										Status = FALSE;
										break;
									}

									/* attempt to handle this chunk */
									SnapshotV3_HandleChunk(pChunk, ChunkSize);

									/* update chunk pointer */
									pChunk = (RIFF_CHUNK *)((unsigned char *)pChunk + ChunkSize + sizeof(RIFF_CHUNK));

									/* update size of data remaining */
									SizeRemaining -= (ChunkSize + sizeof(RIFF_CHUNK));
								}
							}
						}

					}

				}

			}
		}
	}

	return Status;
}

/* calculate the size of the output snapshot based on the passed parameters */
unsigned long Snapshot_CalculateEstimatedOutputSize(SNAPSHOT_OPTIONS *pOptions)
{
	unsigned long SnapshotSize;
	/*    int RamMask = 0; */

	/* fixed size snapshot header */
	SnapshotSize = sizeof(SNAPSHOT_HEADER);

	if (
		/* version 2 */
		(pOptions->Version == 2) ||
		/* uncompressed version 3 */
		((pOptions->Version == 3) && (!pOptions->bCompressed))
		)
	{
		/* base 64K ram; always present */
		SnapshotSize += 64 * 1024;
	}
	else
	{
		/* version 3 compressed */
		SnapshotSize += (64 * 1024) + sizeof(RIFF_CHUNK);
	}


	// TODO: Device ram size
	// Aleste ram size

#if 0
	/* for version 2 only count Dk'Tronics compatible ram */
	/* what about 64 base + 256k silicon? */

	/* extra ram allocated? */
	if (Amstrad_ExtraRam!=NULL)
	{
		/* has 64k ram expansion? */
		if ((CPC_GetRamConfig() & CPC_RAM_CONFIG_64K_RAM)!=0)
		{
			RamMask |= 0x01;
		}

		/* has 256k memory expansion? */
		if ((CPC_GetRamConfig() & CPC_RAM_CONFIG_256K_RAM)!=0)
		{
			RamMask |= 0x0f;
		}

		/* has 256k silicon disk? */
		if ((CPC_GetRamConfig() & CPC_RAM_CONFIG_256K_SILICON_DISK)!=0)
		{
			RamMask |= 0x0f0;
		}

	}

	if (pOptions->Version==2)
	{
		int i;

		for (i=0; i<16; i++)
		{
			if ((RamMask & 1)!=0)
			{
				SnapshotSize += 64*1024;
			}
			RamMask = RamMask>>1;
		}
	}
	else if (pOptions->Version==3)
	{
		int i;
		/* uncompressed size + header for chunk */
		for (i=0; i<16; i++)
		{
			if ((RamMask & 1)!=0)
			{
				SnapshotSize += (64*1024)+sizeof(RIFF_CHUNK);
			}
			RamMask = RamMask>>1;
		}
	}
#endif

	/* TODO: Change version 3 so that it outputs compressed blocks */
	/* and chunked memory */
	if (pOptions->Version == 3)
	{
		/* V3 stuff */

		/* CPC+ hardware? */
		if (CPC_GetHardware() == CPC_HW_CPCPLUS)
		{
			SnapshotSize += SnapshotV3_CPCPlus_CalculateOutputSize();
		}

		/* KCC hardware? */

		/* Aleste hardware? */
	}

	return SnapshotSize;
}



/* fills the supplied pre-allocated buffer with snapshot data. It is the responsibility
of the calling function to allocate and free the data */
size_t Snapshot_GenerateOutputData(unsigned char *pBufferBase, SNAPSHOT_OPTIONS *pOptions)
{
	unsigned char *pBuffer = pBufferBase;
	int             ExtraRAMSize = 0;
	int             SnapshotMemorySize = 0;

	/* initially only write 64K snapshots */
	SnapshotMemorySize = ExtraRAMSize + (64 * 1024);

	{
		int i;
		unsigned char SnapshotHeader[sizeof(SNAPSHOT_HEADER)];
		int CPCType;

		/* clear header */
		memset(&SnapshotHeader, 0, sizeof(SNAPSHOT_HEADER));

		/* setup header */
		memcpy(&SnapshotHeader, SNAPSHOT_HEADER_TEXT, 8);

		memcpy(&SnapshotHeader[0x0e0], SNAPSHOT_EMU_TEXT, 6);

		/* set version */
		SnapshotHeader[0x010] = (unsigned char)pOptions->Version;

		switch (pOptions->Version)
		{
		case 1:
		case 2:
		{
			CPCType = 2;
#if 0
			/* version 2 doesn't recognise KC Compact or CPC+! */
			/* convert my CPC type to Snapshot CPC type index. */
			switch (1)	/*CPC_GetCPCType()) */
			{
			case CPC_TYPE_CPC464:
			{
				CPCType = 0;
			}
			break;

			case CPC_TYPE_CPC664:
			{
				CPCType = 1;
			}
			break;

			case CPC_TYPE_CPC6128:
			{
				CPCType = 2;
			}
			break;

			default:
			{
				/* unknown */
				CPCType = 3;
			}
			break;
			}
#endif
		}
		break;

		default:
		case 3:
		{
			/* version 3 doesn't recognise KC Compact */
			/* convert my CPC type index to Snapshot CPC type index. */
			if (CPC_GetHardware() == CPC_HW_CPCPLUS)
			{
				CPCType = 4;
			}
			else
			{
				CPCType = 2;
			}

#if 0

			/* NOTE: GX4000 not supported by Arnold at this time, so is not defined in this file */
			switch (1)	/*CPC_GetCPCType()) */
			{
			case CPC_TYPE_CPC464:
			{
				CPCType = 0;
			}
			break;

			case CPC_TYPE_CPC664:
			{
				CPCType = 1;
			}
			break;

			case CPC_TYPE_CPC6128:
			{
				CPCType = 2;
			}
			break;

			case CPC_TYPE_6128PLUS:
			{
				CPCType = 4;
			}
			break;

			case CPC_TYPE_464PLUS:
			{
				CPCType = 5;
			}
			break;

			default:
			{
				CPCType = 3;
			}
			break;
			}
#endif
		}
		break;
		}

		/* get cpc type */
		SnapshotHeader[0x06d] = (unsigned char)CPCType;

		/* ***** Z80 CPU ***** */
		{
			int Register;

			char *pRegs = (char *)&SnapshotHeader[0x011];

			Register = CPU_GetReg(CPU_AF);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 0, Register);

			Register = CPU_GetReg(CPU_BC);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 2, Register);

			Register = CPU_GetReg(CPU_DE);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 4, Register);

			Register = CPU_GetReg(CPU_HL);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 6, Register);

			pRegs[8] = (char)CPU_GetReg(CPU_R);
			pRegs[9] = (char)CPU_GetReg(CPU_I);

			/* iff0 */
			pRegs[10] = CPU_GetReg(CPU_IFF1) ? 1 : 0;

			/* iff1 */
			pRegs[11] = CPU_GetReg(CPU_IFF2) ? 1 : 0;

			Register = CPU_GetReg(CPU_IX);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 12, Register);

			Register = CPU_GetReg(CPU_IY);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 14, Register);

			Register = CPU_GetReg(CPU_SP);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 16, Register);

			Register = CPU_GetReg(CPU_PC);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 18, Register);

			Register = CPU_GetReg(CPU_IM);
			pRegs[20] = (char)Register;

			Register = CPU_GetReg(CPU_AF2);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 21, Register);

			Register = CPU_GetReg(CPU_BC2);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 23, Register);

			Register = CPU_GetReg(CPU_DE2);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 25, Register);

			Register = CPU_GetReg(CPU_HL2);

			SNAPSHOT_PUT_REGISTER_PAIR(pRegs, 27, Register);

		}

		switch (CPC_GetHardware())
		{
		case CPC_HW_CPC:
		{
			/**** GATE ARRAY ****/
			/* store colour palette */
			for (i = 0; i < 17; i++)
			{
				SnapshotHeader[0x02f + i] = (unsigned char)((GateArray_GetPaletteColour(i) & 0x01f));
			}

			/* store selected pen */
			SnapshotHeader[0x02e] = (unsigned char)(GateArray_GetSelectedPen());

			/* store multi-configuration */
			SnapshotHeader[0x040] = (unsigned char)((GateArray_GetMultiConfiguration() & 0x01f) | 0x080);
		}
		break;

		case CPC_HW_CPCPLUS:
		{
			/**** GATE ARRAY inside ASIC ****/
			/* store colour palette */
			for (i = 0; i < 17; i++)
			{
				SnapshotHeader[0x02f + i] = (unsigned char)((ASIC_GateArray_GetPaletteColour(i) & 0x01f));
			}

			/* store selected pen */
			SnapshotHeader[0x02e] = (unsigned char)(ASIC_GateArray_GetSelectedPen());

			/* store multi-configuration */
			SnapshotHeader[0x040] = (unsigned char)((ASIC_GateArray_GetMultiConfiguration() & 0x01f) | 0x080);
		}
		break;

		case CPC_HW_KCCOMPACT:
		{
			/* store colour palette */
			for (i = 0; i < 17; i++)
			{
				SnapshotHeader[0x02f + i] = (unsigned char)((KCC_GetColour(i) & 0x01f));
			}

			/* store selected pen */
			SnapshotHeader[0x02e] = (unsigned char)(KCC_GetFN());

			/* store multi-configuration */
			SnapshotHeader[0x040] = (unsigned char)((KCC_GetMF() & 0x01f) | 0x080);

		}
		break;

		case CPC_HW_ALESTE:
		{
			/* store colour palette */
			/* When Aleste is in "MSX mode". This will be RGB */
			for (i = 0; i < 17; i++)
			{
				SnapshotHeader[0x02f + i] = (unsigned char)((Aleste_GetPenColour(i) & 0x01f));
			}

			/* store selected pen */
			SnapshotHeader[0x02e] = (unsigned char)(Aleste_GetCurrentPen());

			/* store multi-configuration; not all bits for aleste! */
			SnapshotHeader[0x040] = (unsigned char)((Aleste_GetMFForSnapshot() & 0x01f) | 0x080);

		}
		break;
		}

		/* store ram configuration */
		SnapshotHeader[0x041] = (unsigned char)(Snapshot_GetRamSelected());

		/**** CRTC ****/
		/* get CRTC status */
		for (i = 0; i < 18; i++)
		{
			SnapshotHeader[0x043 + i] = CRTC_GetRegisterData(i);
		}

		/* select CRTC register */
		SnapshotHeader[0x042] = (unsigned char)(CRTC_GetSelectedRegister() & 0x01f);

		/**** ROM INDEX ****/
		SnapshotHeader[0x055] = (unsigned char)Snapshot_GetRomSelected();


		/**** PSG ****/
		/* get PSG status */
		for (i = 0; i < 16; i++)
		{
			SnapshotHeader[0x05b + i] = (unsigned char)PSG_GetRegisterData(&OnBoardAY, i);
		}

		/* select PSG register */
		SnapshotHeader[0x05a] = (unsigned char)(PSG_GetSelectedRegister(&OnBoardAY) & 0x0f);

		switch (CPC_GetHardware())
		{
		case CPC_HW_CPCPLUS:
		{
			/**** ASIC PPI ****/
			/* get PPI control */
			SnapshotHeader[0x059] = (unsigned char)ASIC_PPI_GetControlForSnapshot();

			/* get port A data */
			SnapshotHeader[0x056] = (unsigned char)ASIC_PPI_GetPortDataForSnapshot(0);
			/* get port B data */
			SnapshotHeader[0x057] = (unsigned char)ASIC_PPI_GetPortDataForSnapshot(1);
			/* get port C data */
			SnapshotHeader[0x058] = (unsigned char)ASIC_PPI_GetPortDataForSnapshot(2);

		}
		break;

		default:
		{
			/* KC Compact has a real 8255 */
			/* CPC has a real 8255 */
			/* Costdown CPC has a real 8255 */
			/* Aleste has a real 8255 */

			/**** PPI ****/
			/* get PPI control */
			SnapshotHeader[0x059] = (unsigned char)PPI_GetControlForSnapshot();

			/* get port A data */
			SnapshotHeader[0x056] = (unsigned char)PPI_GetPortDataForSnapshot(0);
			/* get port B data */
			SnapshotHeader[0x057] = (unsigned char)PPI_GetPortDataForSnapshot(1);
			/* get port C data */
			SnapshotHeader[0x058] = (unsigned char)PPI_GetPortDataForSnapshot(2);
		}
		break;
		}


		/* set memory size */
		SnapshotHeader[0x06b] = (unsigned char)(SnapshotMemorySize >> 10);
		SnapshotHeader[0x06c] = (unsigned char)((SnapshotMemorySize >> 10) >> 8);

		if (
			/* version 2 snapshot */
			(pOptions->Version == 2) ||
			/* uncompressed version 3 snapshot */
			((pOptions->Version == 3) && (!pOptions->bCompressed))
			)
		{
			/* finished header for v2 */
			/* copy it and advance the pointer */

			memcpy(pBuffer, SnapshotHeader, sizeof(SNAPSHOT_HEADER));
			pBuffer += sizeof(SNAPSHOT_HEADER);

			/* copy base 64k out to snapshot */
			memcpy(pBuffer, Z80MemoryBase, (64 << 10));
			pBuffer += (64 << 10);

			/* for version 2 only count Dk'Tronics compatible ram */
#if 0
			/* extra ram allocated? */
			if (Amstrad_ExtraRam!=NULL)
			{
				int RamMask = 0;
				int i;

				/* has 64k ram expansion? */
				if ((CPC_GetRamConfig() & CPC_RAM_CONFIG_64K_RAM)!=0)
				{
					RamMask |= 0x01;
				}

				/* has 256k memory expansion? */
				if ((CPC_GetRamConfig() & CPC_RAM_CONFIG_256K_RAM)!=0)
				{
					RamMask |= 0x0f;
				}

				/* has 256k silicon disk? */
				if ((CPC_GetRamConfig() & CPC_RAM_CONFIG_256K_SILICON_DISK)!=0)
				{
					RamMask |= 0x0f0;
				}

				for (i=0; i<16; i++)
				{
					if ((RamMask & 1)!=0)
					{
						ExtraRAMSize += 64*1024;
					}
					RamMask = RamMask>>1;
				}

				/* copy extra ram */
				memcpy(pBuffer, Amstrad_ExtraRam, ExtraRAMSize);
				pBuffer += ExtraRAMSize;
			}
#endif
		}
		else
		{
			/* TODO: Move version 3 into it's own function */
			if (pOptions->Version == 3)
			{
				/* poke V3 stuff into header */
				unsigned long CRTC_Flags;

				/* put V3 stuff into header */
				CRTC_INTERNAL_STATE *pCRTC_State;

				pCRTC_State = CRTC_GetInternalState();

				if (FDI_GetMotorState())
				{
					SnapshotHeader[0x09c] = 1;
				}
				else
				{
					SnapshotHeader[0x09c] = 0;
				}

				/* last byte written to printer port */
				/* TODO: Check! */
				SnapshotHeader[0x0a1] = Printer_Get7BitData();
				if (Printer_GetStrobeOutputState())
				{
					SnapshotHeader[0x0a1] |= 0x080;
				}

				/* crtc type */
				SnapshotHeader[0x0a4] = CPC_GetCRTCType();

				/* crtc internal state */
				SnapshotHeader[0x0a9] = pCRTC_State->HCount;
				SnapshotHeader[0x0ab] = pCRTC_State->LineCounter;
				SnapshotHeader[0x0ac] = pCRTC_State->RasterCounter;
				SnapshotHeader[0x0ad] = pCRTC_State->VertAdjustCount;

				CRTC_Flags = 0;

				if (pCRTC_State->CRTC_Flags & CRTC_VS_FLAG)
				{
					SnapshotHeader[0x0af] = pCRTC_State->VerticalSyncCount & 0x0f;
					CRTC_Flags |= SNAPSHOT_CRTC_FLAGS_VSYNC_STATE;
				}

				if (pCRTC_State->CRTC_Flags & CRTC_HS_FLAG)
				{
					SnapshotHeader[0x0ae] = pCRTC_State->HorizontalSyncCount & 0x0f;
					CRTC_Flags |= SNAPSHOT_CRTC_FLAGS_HSYNC_STATE;
				}

				if (pCRTC_State->CRTC_Flags & CRTC_VADJ_FLAG)
				{
					SnapshotHeader[0x0ad] = pCRTC_State->VertAdjustCount & 0x01f;
					CRTC_Flags |= SNAPSHOT_CRTC_FLAGS_VADJ_STATE;
				}


				SnapshotHeader[0x0b0] = CRTC_Flags & 0x0ff;
				SnapshotHeader[0x0b1] = (CRTC_Flags >> 8) & 0x0ff;


				if (CPU_GetINTState())
				{
					SnapshotHeader[0x0b4] = 1;
				}
				else
				{
					SnapshotHeader[0x0b4] = 0;
				}

				switch (CPC_GetHardware())
				{
				case CPC_HW_CPC:
				{
					int nVBlankCycle = 0;
					if (GateArray_GetVBlankActive())
					{
						nVBlankCycle = GateArray_GetVBlankCount();
					}
					if (nVBlankCycle > 2)
					{
						nVBlankCycle = 0;
					}

					SnapshotHeader[0x0b2] = nVBlankCycle;
					SnapshotHeader[0x0b3] = GateArray_GetInterruptLineCount();
				}
				break;

				case CPC_HW_CPCPLUS:
				{
					int nVBlankCycle = 0;
					if (ASIC_GateArray_GetVBlankActive())
					{
						nVBlankCycle = ASIC_GateArray_GetVBlankCount();
					}
					if (nVBlankCycle > 2)
					{
						nVBlankCycle = 0;
					}

					SnapshotHeader[0x0b2] = nVBlankCycle;
					SnapshotHeader[0x0b3] = ASIC_GateArray_GetInterruptLineCount();
				}
				break;

				case CPC_HW_KCCOMPACT:
				{
					/* interrupt line count is set inside z8536 */
					SnapshotHeader[0x0b2] = 0;
					SnapshotHeader[0x0b3] = 0;
				}
				break;

				case CPC_HW_ALESTE:
				{
					SnapshotHeader[0x0b2] = 0;
					SnapshotHeader[0x0b3] = 0;
				}
				break;
				}

				/* finished header for v3 */
				/* copy it and advance the pointer */
				memcpy(pBuffer, SnapshotHeader, sizeof(SNAPSHOT_HEADER));
				pBuffer += sizeof(SNAPSHOT_HEADER);

				/* if compressed, now write memory */
				if (pOptions->bCompressed)
				{
					/* write out dk'tronics compatible ram */
					pBuffer = SnapshotV3_Memory_WriteChunk(pBuffer);
				}


				/* CPC+ hardware? */
				if (CPC_GetHardware() == CPC_HW_CPCPLUS)
				{
					pBuffer = SnapshotV3_CPCPlus_WriteChunk(pBuffer);
				}

				/* KCC hardware? */


				/* Aleste hardware? */

				/*	pBuffer = SnapshotV3_DiscFiles_WriteChunk(pBuffer); */
			}
		}
	}

	return pBuffer - pBufferBase;
}
