/*
*  Arnold emulator (c) Copyright, Kevin Thacker 1995-2015
*
*  This file is part of the Arnold emulator source code distribution.
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include "cpc.h"
#include "EmuDevice.h"
#include "mouse.h"
#include "ata.h"

enum
{
    SYMBIMOUSE_MODE_NONE = 0,
    SYMBIMOUSE_MODE_X_MOVEMENT,
    SYMBIMOUSE_MODE_Y_MOVEMENT,
    SYMBIMOUSE_MODE_BUTTONS
};


typedef struct 
{
	unsigned char Registers[128];
	unsigned char Address;
	unsigned char UpdateStatus;
	unsigned char InterruptFlags;
} ds12887;
	
ds12887 symbiface_rtc;
ata_device symbiface_ata;

void ds12887_reset(ds12887 *ds)
{
	ds->UpdateStatus  = 0;
	/* reset periodic interrupt enable,
		reset alaram interrupt enable,
		reset update ended interrupt enable,
		square wave enable reset */
	ds->Registers[0x0b] &=~((1<<6)|(1<<5)|(1<<4)|(1<<3));

	ds->InterruptFlags = 0;
}

void ds12887_update_ints(ds12887 *ds)
{
	ds->InterruptFlags &=~(1<<7);
	
	/* if PF=PIE=1, AF=AIE=1, UF=UIE=1 set IRQF */
	if (ds->InterruptFlags & ds->Registers[0x0b] & ((1<<6)|(1<<5)|(1<<4)))
	{
		ds->InterruptFlags |= (1<<7);
	}
}

unsigned char ds12887_read_data(ds12887 *ds)
{
	//printf("Read ds12887 Addr %02x\n", ds->Address);

	ds->Registers[0]++;
	
	if (ds->Address==0x0c)
		return ds->InterruptFlags;
	if (ds->Address==0x0d)
		return (1<<7);	/* indicates valid ram and time */
	
	{
		unsigned char Data = ds->Registers[ds->Address];
		if (ds->Address==0x0a)
		{
			Data = Data & 0x07f;
			Data = Data|ds->UpdateStatus;
		}
	
		return Data;
	}
}

void ds12887_write_addr(ds12887 *ds, unsigned char Data)
{
	ds->Address = Data &0x07f;
	//printf("Select ds12887 Addr %02x\n", ds->Address);
}


void ds12887_write_data(ds12887 *ds, unsigned char Data)
{
	//printf("Write ds12887 Addr %02x Data %02x\n", ds->Address, Data);
	ds->Registers[ds->Address] = Data;
}

typedef struct 
{
	BOOL bStoredMouseValid;
	BOOL bStoreMouse;
	signed int StoredDeltaX;
	signed int StoredDeltaY;
	int StoredMouseX;
	int StoredMouseY;
	int StoredMouseButtons;
	int StoredHDData;
	int MouseX;
	int MouseY;
	int SymbifaceReport;
} symbiface2;

static symbiface2 sym2;

#define SYMBIFACE_REPORT_X (1<<0)
#define SYMBIFACE_REPORT_Y (1<<1)
#define SYMBIFACE_REPORT_BUTTONS (1<<2)


#define SYMBIMOUSE_BUTTON_SCROLL_WHEEL_STATE (1<<5)

#define SYMBIMOUSE_BUTTON_LEFT (1<<0)
#define SYMBIMOUSE_BUTTON_RIGHT (1<<1)
#define SYMBIMOUSE_BUTTON_MIDDLE (1<<2)
#define SYMBIMOUSE_BUTTON_FORWARD (1<<3)
#define SYMBIMOUSE_BUTTON_BACKWARD (1<<4)

/*
// fd06-fd0f for ide

// fd10 is mouse (write only)
// fd11 enable/disable roms
// fd14 is RTC data (read/write)
// fd15 is RTC register (write only)
// fd17 show rom or not
*/

char intto6bitsigned(int x)
{
    char ax5 = ((char)abs(x)) & 0x1f;
    if (x<0)
    {
        return ((ax5^0x3f)+1);
    }
    else
    {
        return (ax5);
    }
}

Z80_BYTE SymbifaceRead(Z80_WORD Port)
{
//	//printf("Symbiface read %04x\n", Port);
    switch (Port & 0x01f)
    {


		case 6:
			return ata_read(&symbiface_ata,ATA_ALTERNATE_STATUS_CS | ATA_ALTERNATE_STATUS_DA);

		case 7:
		{


		}
		break;

		/* TODO read 16-bit data. symbiface reads 8-bit at a time */
		/* IDE data register */
		case 0x8:
			return ata_read(&symbiface_ata,ATA_DATA_REGISTER_DA | ATA_DATA_REGISTER_CS) & 0x0ff;

		/* IDE Error register */
		case 0x9:
			return ata_read(&symbiface_ata,ATA_ERROR_DA | ATA_ERROR_CS) & 0x0ff;
			
		/* IDE Sector count */
		case 0x0A:
			return ata_read(&symbiface_ata,ATA_SECTOR_COUNT_DA | ATA_SECTOR_COUNT_CS);

		/* IDE Sector number */
		case 0x0b:
			return ata_read(&symbiface_ata,ATA_SECTOR_NUMBER_DA | ATA_SECTOR_NUMBER_CS);

		/* IDE Cylinder number low */
		case 0x0c:
			return ata_read(&symbiface_ata,ATA_CYLINDER_NUMBER_LOW_DA | ATA_CYLINDER_NUMBER_LOW_CS);
		
		/* IDE Cylinder number high */
		case 0x0d:
			return ata_read(&symbiface_ata,ATA_CYLINDER_NUMBER_HIGH_DA | ATA_CYLINDER_NUMBER_HIGH_CS);

		/* IDE head */
		case 0x0e:
			return ata_read(&symbiface_ata,ATA_HEAD_DA | ATA_HEAD_CS);

		/* IDE Status */
		case 0x0f:
			return ata_read(&symbiface_ata,ATA_STATUS_DA | ATA_STATUS_CS);

				case 0x010:
	{
		Z80_BYTE Data = 0x00;
		/* PS/2 protocol:

		mouse can send packet
		or host can choose to read packet

		symbiface must store it; or add it to it's values
		because it doesn't know when CPC will read it. CPC may choose to read
		it at any time.

		when read, it returns the value and then resets it.

		then it can know if value should be passed to CPC or not.

		*/
		if (sym2.SymbifaceReport == 0)
		{
			signed int MouseX = Mouse_GetX();
			signed int MouseY = Mouse_GetY();
			int MouseButtons = Mouse_GetButtons();

			if ( !sym2.bStoredMouseValid )
			{
				sym2.bStoredMouseValid = TRUE;
				sym2.StoredMouseX = MouseX;
				sym2.StoredMouseY = MouseY;
				sym2.StoredMouseButtons = MouseButtons;
			}


			sym2.StoredDeltaX = MouseX - sym2.StoredMouseX;
			sym2.StoredDeltaY = -( MouseY - sym2.StoredMouseY );
			sym2.StoredMouseX = MouseX;
			sym2.StoredMouseY = MouseY;
			if ( sym2.StoredDeltaX )
			{
				sym2.SymbifaceReport |= SYMBIFACE_REPORT_X;
			}
			if ( sym2.StoredDeltaY )
			{
				sym2.SymbifaceReport |= SYMBIFACE_REPORT_Y;
			}
			if ( sym2.StoredMouseButtons^MouseButtons )
			{
				sym2.StoredMouseButtons = MouseButtons;
				sym2.SymbifaceReport |= SYMBIFACE_REPORT_BUTTONS;
			}
		}

		if ( sym2.SymbifaceReport & SYMBIFACE_REPORT_X )
		{
			Data = ( SYMBIMOUSE_MODE_X_MOVEMENT << 6 ) | intto6bitsigned( sym2.StoredDeltaX );
			sym2.SymbifaceReport &= ~SYMBIFACE_REPORT_X;
		}
		else if ( sym2.SymbifaceReport & SYMBIFACE_REPORT_Y )
		{
			Data = ( SYMBIMOUSE_MODE_Y_MOVEMENT << 6 ) | intto6bitsigned( sym2.StoredDeltaY );
			sym2.SymbifaceReport &= ~SYMBIFACE_REPORT_Y;
		}
		else if ( sym2.SymbifaceReport & SYMBIFACE_REPORT_BUTTONS )
		{
			Data = ( SYMBIMOUSE_MODE_BUTTONS << 6 ) | sym2.StoredMouseButtons;
			sym2.SymbifaceReport &= ~SYMBIFACE_REPORT_BUTTONS;
		}

		return Data;
	}
    break;

				case 0x014:
				{
					/* rtc data */
					unsigned char Data = ds12887_read_data(&symbiface_rtc);
					//printf("Read ds12887 %02x\n", Data);

				}
				break;


		default:
			//printf("Symbiface read not supported %04x", Port);
			break;

    }

    return 0x0ff;
}


void SymbifaceWrite(Z80_WORD Port, Z80_BYTE Data)
{
//	//printf("Symbiface write %04x %02x\n", Port, Data);

	switch (Port & 0x01f)
	{
	
	// TODO: Verify 8-bit or 16-bit reading
	case 8:
	{
		ata_write(&symbiface_ata,ATA_DATA_REGISTER_DA | ATA_DATA_REGISTER_CS, Data & 0x0ff);
	}
	break;

	case 0x9:
	{
		ata_write(&symbiface_ata,ATA_FEATURES_DA | ATA_FEATURES_CS, Data);
	}
	break;

	case 0x0A:
	{
		ata_write(&symbiface_ata,ATA_SECTOR_COUNT_DA | ATA_SECTOR_COUNT_CS, Data);
	}
	break;

	case 0x0b:
	{
		ata_write(&symbiface_ata,ATA_SECTOR_NUMBER_DA | ATA_SECTOR_NUMBER_CS, Data);
	}
	break;

	case 0x0c:
	{
		ata_write(&symbiface_ata,ATA_CYLINDER_NUMBER_LOW_DA | ATA_CYLINDER_NUMBER_LOW_CS, Data);
	}
	break;

	case 0x0d:
	{
		ata_write(&symbiface_ata,ATA_CYLINDER_NUMBER_HIGH_DA | ATA_CYLINDER_NUMBER_HIGH_CS, Data);
	}
	break;

	case 0x0e:
	{
		ata_write(&symbiface_ata,ATA_HEAD_DA | ATA_HEAD_CS, Data);
	}
	break;

	case 0x0f:
	{
		ata_write(&symbiface_ata,ATA_COMMAND_DA | ATA_COMMAND_CS, Data);
	}
	break;

	case 0x014:
	{
		ds12887_write_data(&symbiface_rtc, Data);
	}
	break;
	
	case 0x015:
	{
		ds12887_write_addr(&symbiface_rtc, Data);
	}
	break;
	
	
	default:
	{

		//printf("Symbiface write not supported %04x %02x", Port, Data);
	}
	break;

	}
}

void Symbiface_Reset(void)
{
	ds12887_reset(&symbiface_rtc);
	ata_reset(&symbiface_ata);
}


static CPCPortRead symbifacePortRead=
{
    0x0ff00, /* and */
    0x0fd00, /* cmp */
    SymbifaceRead
};

static CPCPortWrite symbifacePortWrite=
{
    0x0ff00, /* and */
    0x0fd00, /* cmp */
    SymbifaceWrite
};


void Symbiface_InitDevice(void)
{
	ata_init(&symbiface_ata, 20 * 1024 * 1024);
}

void Symbiface_FinishDevice(void)
{
	ata_finish(&symbiface_ata);
}

static EmuDevice SymbifaceDevice=
{
	Symbiface_InitDevice,
	Symbiface_FinishDevice,
	"SYMBIFACE",
	"Symbiface",
	CONNECTION_EXPANSION,   /* connected to expansion */
	0,
  1,                /* 1 read port */
  &symbifacePortRead,
  1,                    /* 1 write ports */
  &symbifacePortWrite, 
  0,                /* no memory read*/
  NULL,
  0,                /* no memory write */
  NULL,
  Symbiface_Reset, /* reset function */
  NULL, /* memory rethink */
  Symbiface_Reset, /* power function*/
	0,      /* no switches*/
	NULL,
    0,                      /* no buttons */
    NULL,
    0,                      /* no onboard roms */
    NULL,
      NULL,                   /* no cursor function */
      NULL,                 /* no generic roms */
	  NULL,
	  NULL,
	  0,
	  NULL,
	  NULL, /* sound */
	  NULL, /* lpen */
	  NULL, /* reti */
	  NULL, /* ack maskable interrupt */
	  NULL, /* dkram data */
	  NULL, /* device ram */
	  NULL, /* device backup */

};

void Symbiface_Init(void)
{
	memset( &sym2, 0, sizeof( symbiface2 ) );
	sym2.bStoredMouseValid = FALSE;
	sym2.bStoreMouse = TRUE;
	sym2.StoredDeltaX = 0;
	sym2.StoredDeltaY = 0;
	sym2.StoredMouseX = 0;
	sym2.StoredMouseY = 0;
	sym2.StoredMouseButtons = 0;
	sym2.StoredHDData = 0;
	sym2.MouseX = 0;
	sym2.MouseY = 0;
	sym2.SymbifaceReport = 0;
	RegisterDevice(&SymbifaceDevice);
}
